/*
 * Copyright 2007 Dmitry A.Steklenev <glass@ptv.ru>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *    1. Redistributions of source code must retain the above copyright notice,
 *       this list of conditions and the following disclaimer.
 *
 *    2. Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *
 *    3. The name of the author may not be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#define  INCL_DOS
#define  INCL_WIN
#include <os2.h>
#include <malloc.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <process.h>
#include <errno.h>

#include <vorbis/vorbisfile.h>
#include <decoder_plug.h>
#include <plugin.h>
#include <utilfct.h>
#include <snprintf.h>

#include "oggplay.h"
#include "vcedit.h"

/* Changes the current file position to a new location within the file.
   Returns 0 if it successfully moves the pointer. A nonzero return
   value indicates an error. On devices that cannot seek the return
   value is nonzero. */
static int DLLENTRY
vio_seek( void* w, ogg_int64_t offset, int whence )
{
  int    pos = 0;
  XFILE* x   = ((DECODER_STRUCT*)w)->file;

  if( xio_can_seek(x) < XIO_CAN_SEEK_FAST ) {
    return -1;
  }

  switch( whence )
  {
    case SEEK_SET: pos = offset; break;
    case SEEK_CUR: pos = xio_ftell(x) + offset; break;
    case SEEK_END: pos = xio_fsize(x) + offset; break;
    default:
      return -1;
  }

  if( xio_fseek( x, pos, XIO_SEEK_SET ) == 0 ) {
    return pos;
  } else {
    return -1;
  }
}

/* Finds the current position of the file. Returns the current file
   position. On error, returns -1L and sets errno to a nonzero value. */
static long DLLENTRY
vio_tell( void* w ) {
  return xio_ftell(((DECODER_STRUCT*)w)->file );
}

/* Reads up to count items of size length from the input file and
   stores them in the given buffer. The position in the file increases
   by the number of bytes read. Returns the number of full items
   successfully read, which can be less than count if an error occurs
   or if the end-of-file is met before reaching count. */
static size_t DLLENTRY
vio_read( void* ptr, size_t size, size_t count, void* w ) {
  return xio_fread( ptr, size, count, ((DECODER_STRUCT*)w)->file );
}

/* Closes a pointed file. Returns 0 if it successfully closes
   the file, or -1 if any errors were detected. */
static int DLLENTRY
vio_close( void* w )
{
  int rc = xio_fclose(((DECODER_STRUCT*)w)->file );
  ((DECODER_STRUCT*)w)->file = NULL;
  return rc;
}

/* Open file. Returns 0 if it successfully opens the file.
   A nonzero return value indicates an error. */
static ULONG
ogg_open( DECODER_STRUCT* w )
{
  ov_callbacks callbacks;
  int rc;

  if(( w->file = xio_fopen( w->filename, "rb" )) == NULL ) {
    return xio_errno();
  }

  callbacks.read_func  = vio_read;
  callbacks.seek_func  = vio_seek;
  callbacks.tell_func  = vio_tell;
  callbacks.close_func = vio_close;

  // The ov_open_callbaks() function performs full stream detection and machine
  // initialization. If it returns 0, the stream *is* Vorbis and we're
  // fully ready to decode.
  if(( rc = ov_open_callbacks( w, &w->vrbfile, NULL, 0, callbacks )) < 0 ) {
    xio_fclose( w->file );
    return PLUGIN_NO_PLAY;
  }

  return PLUGIN_OK;
}

/* Reads up to count pcm bytes from the Ogg Vorbis stream and
   stores them in the given buffer. */
static int
ogg_read( DECODER_STRUCT* w, char* buffer, int count )
{
  int done;
  int read    =  0;
  int section = -1;
  int resync  =  0;

  while( read < count )
  {
    done = ov_read( &w->vrbfile, buffer + read, count - read, 0, 2, 1, &section );

    if( done == OV_HOLE ) {
      if( resync++ < MAXRESYNC ) {
        continue;
      }
    }
    if( done <= 0 ) {
      break;
    } else {
      read += done;
    }
  }

  return read;
}

/* Decoding thread. */
static void TFNENTRY
decoder_thread( void* arg )
{
  ULONG resetcount;
  ULONG markerpos;
  ULONG rc;

  DECODER_STRUCT* w = (DECODER_STRUCT*)arg;
  vorbis_info*    vrbinfo;
  vorbis_comment* comment;

  char* buffer = NULL;
  char  last_title[512] = "";
  char* curr_title;

  w->frew = FALSE;
  w->ffwd = FALSE;

  if(( rc = ogg_open( w )) != 0 ) {
    if( w->error_display )
    {
      char errorbuf[1024];

      strlcpy( errorbuf, "Unable open file:\n", sizeof( errorbuf ));
      strlcat( errorbuf, w->filename, sizeof( errorbuf ));
      strlcat( errorbuf, "\n", sizeof( errorbuf ));
      if( rc != PLUGIN_NO_PLAY ) {
        strlcat( errorbuf, xio_strerror( xio_errno()), sizeof( errorbuf ));
      } else {
        strlcat( errorbuf, "Unsupported format of the file.", sizeof( errorbuf ));
      }

      w->error_display( errorbuf );
    }
    WinPostMsg( w->hwnd, WM_PLAYERROR, 0, 0 );
    goto end;
  }

  // After opening a new file we so are in its beginning.
  if( w->jumptopos == 0 ) {
      w->jumptopos = -1;
  }

  if(!( buffer = (unsigned char*)malloc( w->audio_buffersize ))) {
    WinPostMsg( w->hwnd, WM_PLAYERROR, 0, 0 );
    goto end;
  }

  vrbinfo = ov_info( &w->vrbfile, -1 );
  comment = ov_comment( &w->vrbfile, -1 );
  w->songlength = ov_time_total( &w->vrbfile, -1 ) * 1000.0;

  w->output_format.size       = sizeof( w->output_format );
  w->output_format.format     = WAVE_FORMAT_PCM;
  w->output_format.bits       = 16;
  w->output_format.channels   = vrbinfo->channels;
  w->output_format.samplerate = vrbinfo->rate;

  if( w->metadata_buff && w->metadata_size ) {
    xio_set_observer( w->file, w->hwnd, w->metadata_buff, w->metadata_size );
  }

  for(;;)
  {
    DosWaitEventSem ( w->play, SEM_INDEFINITE_WAIT );
    DosResetEventSem( w->play, &resetcount );

    if( w->stop ) {
      break;
    }

    w->status = DECODER_PLAYING;

    while( !w->stop )
    {
      int read;
      int write;

      if( w->resumepos >  0 ) {
        ov_pcm_seek( &w->vrbfile, w->resumepos );
        w->resumepos = -1;
      }

      if( w->jumptopos >= 0 )
      {
        ov_time_seek( &w->vrbfile, w->jumptopos / 1000.0 );
        w->jumptopos = -1;
        WinPostMsg( w->hwnd, WM_SEEKSTOP, 0, 0 );
      }

      if( w->frew ) {
        /* ie.: ~1/2 second */
        double pos = ov_time_tell( &w->vrbfile ) - 0.500;

        if( pos < 0 ) {
          break;
        } else {
          ov_time_seek( &w->vrbfile, pos );
        }
      }

      if( w->ffwd ) {
        ov_time_seek( &w->vrbfile, ov_time_tell( &w->vrbfile ) + 0.500 );
      }

      markerpos = 1000.0 * ov_time_tell( &w->vrbfile );
      read = ogg_read( w, buffer, w->audio_buffersize );

      if( read <= 0 ) {
        break;
      }

      write = w->output_play_samples( w->a, &w->output_format, buffer, read, markerpos );
      w->playedpos = ov_pcm_tell( &w->vrbfile );

      if( write != read ) {
        WinPostMsg( w->hwnd, WM_PLAYERROR, 0, 0 );
        break;
      }

      if(( curr_title = vorbis_comment_query( comment, "TITLE", 0 )) != NULL &&
           strcmp( curr_title, last_title ) != 0 && w->metadata_buff && w->metadata_size )
      {
        char stream_title[128];

        strlcpy( last_title, curr_title, sizeof( last_title ));
        ch_convert( CH_UTF_8, last_title, CH_DEFAULT, stream_title, sizeof( stream_title ));
        snprintf( w->metadata_buff, w->metadata_size, "StreamTitle='%s';", stream_title );
        WinPostMsg( w->hwnd, WM_METADATA, MPFROMP( w->metadata_buff ), 0 );
      }
    }
    WinPostMsg( w->hwnd, WM_PLAYSTOP, 0, 0 );
    w->status = DECODER_STOPPED;
  }

end:

  DosRequestMutexSem( w->mutex, SEM_INDEFINITE_WAIT );
  ov_clear( &w->vrbfile );
  free( buffer );

  w->status = DECODER_STOPPED;
  DosReleaseMutexSem( w->mutex );
  _endthread();
}

/* Init function is called when PM123 needs the specified decoder to play
   the stream demanded by the user. */
int DLLENTRY
decoder_init( void** returnw )
{
  DECODER_STRUCT* w = (DECODER_STRUCT*)calloc( sizeof(*w), 1 );
  *returnw = w;

  DosCreateEventSem( NULL, &w->play,  0, FALSE );
  DosCreateMutexSem( NULL, &w->mutex, 0, FALSE );

  w->decodertid = -1;
  w->status = DECODER_STOPPED;
  w->file = NULL;
  return PLUGIN_OK;
}

/* Uninit function is called when another decoder than this is needed. */
BOOL DLLENTRY
decoder_uninit( void* arg )
{
  DECODER_STRUCT* w = (DECODER_STRUCT*)arg;

  decoder_command ( w, DECODER_STOP, NULL );
  DosCloseEventSem( w->play  );
  DosCloseMutexSem( w->mutex );
  free( w );
  return TRUE;
}

/* There is a lot of commands to implement for this function. Parameters
   needed for each of the are described in the definition of the structure
   in the decoder_plug.h file. */
ULONG DLLENTRY
decoder_command( void* arg, ULONG msg, DECODER_PARAMS* info )
{
  DECODER_STRUCT* w = arg;

  switch( msg )
  {
    case DECODER_SETUP:
      w->output_play_samples = info->output_play_samples;
      w->hwnd                = info->hwnd;
      w->error_display       = info->error_display;
      w->info_display        = info->info_display;
      w->a                   = info->a;
      w->audio_buffersize    = info->audio_buffersize;
      w->metadata_buff       = info->metadata_buffer;
      w->metadata_buff[0]    = 0;
      w->metadata_size       = info->metadata_size;
      break;

    case DECODER_PLAY:
    {
      DosRequestMutexSem( w->mutex, SEM_INDEFINITE_WAIT );

      if( w->decodertid != -1 ) {
        DosReleaseMutexSem( w->mutex );
        if( w->status == DECODER_STOPPED ) {
          decoder_command( w, DECODER_STOP, NULL );
          DosRequestMutexSem( w->mutex, SEM_INDEFINITE_WAIT );
        } else {
          return PLUGIN_GO_ALREADY;
        }
      }

      strlcpy( w->filename, info->filename, sizeof( w->filename ));

      w->jumptopos  = info->jumpto;
      w->resumepos  = -1;
      w->status     = DECODER_STARTING;
      w->stop       = FALSE;
      w->songlength = 0;
      w->decodertid = _beginthread( decoder_thread, 0, 65535, (void*)w );

      DosReleaseMutexSem( w->mutex );
      DosPostEventSem   ( w->play  );
      break;
    }

    case DECODER_RESUME:
    {
      DosRequestMutexSem( w->mutex, SEM_INDEFINITE_WAIT );

      if( w->decodertid != -1 ) {
        DosReleaseMutexSem( w->mutex );
        return PLUGIN_GO_ALREADY;
      }
      if( w->status != DECODER_PAUSED ) {
        DosReleaseMutexSem( w->mutex );
        return PLUGIN_GO_FAILED;
      }

      w->status     = DECODER_STARTING;
      w->stop       = FALSE;
      w->decodertid = _beginthread( decoder_thread, 0, 65535, (void*)w );

      DosReleaseMutexSem( w->mutex );
      DosPostEventSem   ( w->play  );
      break;
    }

    case DECODER_SUSPEND:
    {
      DosRequestMutexSem( w->mutex, SEM_INDEFINITE_WAIT );

      if( w->decodertid == -1 ) {
        DosReleaseMutexSem( w->mutex );
        return PLUGIN_GO_ALREADY;
      }
      if( xio_can_seek( w->file ) != XIO_CAN_SEEK_FAST ) {
        DosReleaseMutexSem( w->mutex );
        return PLUGIN_UNSUPPORTED;
      }

      w->stop = TRUE;

      if( w->file ) {
        xio_fabort( w->file );
      }

      DosReleaseMutexSem( w->mutex );
      DosPostEventSem( w->play  );
      wait_thread( w->decodertid, 5000 );

      w->decodertid = -1;
      w->resumepos  = w->playedpos;
      w->status     = DECODER_PAUSED;
      w->file       = NULL;
      break;
    }

    case DECODER_STOP:
    {
      DosRequestMutexSem( w->mutex, SEM_INDEFINITE_WAIT );

      if( w->decodertid == -1 ) {
        DosReleaseMutexSem( w->mutex );
        return PLUGIN_GO_ALREADY;
      }

      w->stop = TRUE;

      if( w->file ) {
        xio_fabort( w->file );
      }

      DosReleaseMutexSem( w->mutex );
      DosPostEventSem( w->play  );
      wait_thread( w->decodertid, 5000 );

      w->decodertid = -1;
      w->file = NULL;
      break;
    }

    case DECODER_FFWD:
      if( info->ffwd ) {
        DosRequestMutexSem( w->mutex, SEM_INDEFINITE_WAIT );
        if( w->decodertid == -1 || xio_can_seek( w->file ) != XIO_CAN_SEEK_FAST ) {
          DosReleaseMutexSem( w->mutex );
          return PLUGIN_UNSUPPORTED;
        }
        DosReleaseMutexSem( w->mutex );
      }
      w->ffwd = info->ffwd;
      break;

    case DECODER_REW:
      if( info->rew ) {
        DosRequestMutexSem( w->mutex, SEM_INDEFINITE_WAIT );
        if( w->decodertid == -1 || xio_can_seek( w->file ) != XIO_CAN_SEEK_FAST ) {
          DosReleaseMutexSem( w->mutex );
          return PLUGIN_UNSUPPORTED;
        }
        DosReleaseMutexSem( w->mutex );
      }
      w->frew = info->rew;
      break;

    case DECODER_JUMPTO:
      w->jumptopos = info->jumpto;
      if( w->status == DECODER_STOPPED ) {
        DosPostEventSem( w->play );
      }
      break;

    default:
      return PLUGIN_UNSUPPORTED;
   }

   return PLUGIN_OK;
}

/* Returns current status of the decoder. */
ULONG DLLENTRY
decoder_status( void* arg ) {
  return ((DECODER_STRUCT*)arg)->status;
}

/* Returns number of milliseconds the stream lasts. */
ULONG DLLENTRY
decoder_length( void* arg ) {
  return ((DECODER_STRUCT*)arg)->songlength;
}

static void
copy_string( vorbis_comment* comment, char* target, char* type, int size )
{
  const char* string;

  if(( string = vorbis_comment_query( comment, type, 0 )) != NULL ) {
    strlcpy( target, string, size );
  } else if( size ) {
    *target = 0;
  }
}

/* Returns information about specified file. */
ULONG DLLENTRY
decoder_fileinfo( char* filename, DECODER_INFO* info )
{
  DECODER_STRUCT w = { 0 };
  ULONG rc;

  strlcpy( w.filename, filename, sizeof( w.filename ));
  if(( rc = ogg_open( &w )) == PLUGIN_OK )
  {
    vorbis_info*    vrbinfo = ov_info( &w.vrbfile, -1 );
    vorbis_comment* comment = ov_comment( &w.vrbfile, -1 );

    if( !vrbinfo ) {
      ov_clear( &w.vrbfile );
      return PLUGIN_NO_READ;
    }

    info->format.size       = sizeof( info->format );
    info->format.format     = WAVE_FORMAT_PCM;
    info->format.bits       = 16;
    info->format.channels   = vrbinfo->channels;
    info->format.samplerate = vrbinfo->rate;
    info->mode              = info->format.channels == 1 ? 3 : 0;
    info->bitrate           = ov_bitrate( &w.vrbfile, -1 ) / 1000;
    info->songlength        = ov_time_total( &w.vrbfile, -1 ) * 1000;

    if( info->size >= INFO_SIZE_2 ) {
      info->filesize = xio_fsize( w.file );
    }

    if( info->songlength <= 0 ) {
        info->songlength = -1;
    }

    snprintf( info->tech_info, sizeof( info->tech_info ), "%u kbs, %.1f kHz, %s",
              info->bitrate, ( info->format.samplerate / 1000.0 ),
              info->format.channels == 1 ? "Mono" : "Stereo" );

    if( comment )
    {
      const char* string;

      copy_string( comment, info->artist,  "ARTIST",  sizeof( info->artist  ));
      copy_string( comment, info->album,   "ALBUM",   sizeof( info->album   ));
      copy_string( comment, info->title,   "TITLE",   sizeof( info->title   ));
      copy_string( comment, info->genre,   "GENRE",   sizeof( info->genre   ));
      copy_string( comment, info->year,    "YEAR",    sizeof( info->year    ));
      copy_string( comment, info->comment, "COMMENT", sizeof( info->comment ));

      if( !*info->year ) {
        copy_string( comment, info->year, "DATE", sizeof( info->year ));
      }

      if( info->size >= INFO_SIZE_2 )
      {
        copy_string( comment, info->copyright, "COPYRIGHT", sizeof( info->copyright ));

        if(( string = vorbis_comment_query( comment, "TRACKNUMBER", 0 )) != NULL )
        {
          int track = atoi( string );
          if( track ) {
            sprintf( info->track, "%02d", track );
          }
        }

        info->haveinfo = DECODER_HAVE_TITLE    |
                         DECODER_HAVE_ARTIST   |
                         DECODER_HAVE_ALBUM    |
                         DECODER_HAVE_YEAR     |
                         DECODER_HAVE_GENRE    |
                         DECODER_HAVE_TRACK    |
                         DECODER_HAVE_COMMENT  |
                         DECODER_HAVE_COPYRIGHT;

        info->saveinfo = xio_can_seek( w.file ) == XIO_CAN_SEEK_FAST;
        info->codepage = CH_UTF_8;
      }
    }
    ov_clear( &w.vrbfile );
  }

  return rc;
}

/* It is called if it is necessary to change the meta information
   for the specified file. */
ULONG DLLENTRY
decoder_saveinfo( char* filename, DECODER_INFO* info, char* savename )
{
  FILE* file;
  FILE* save;
  int   rc = PLUGIN_OK_COPY;

  vcedit_state*   state = NULL;
  vorbis_comment* vc;

  for(;;)
  {
    if(( file = fopen( filename, "rb" )) == NULL ) {
      rc = errno;
      break;
    }
    if(( save = fopen( savename, "wb" )) == NULL ) {
      rc = errno;
      break;
    }

    state = vcedit_new_state();

    if( vcedit_open_callbacks( state, file, (vcedit_read_func )fread,
                                            (vcedit_write_func)fwrite ) < 0 )
    {
      rc = EINVAL;
      break;
    }

    vc = vcedit_comments( state );
    vorbis_comment_clear( vc );
    vorbis_comment_init ( vc );

    vorbis_comment_add_tag( vc, "ARTIST",      info->artist  );
    vorbis_comment_add_tag( vc, "ALBUM",       info->album   );
    vorbis_comment_add_tag( vc, "TITLE",       info->title   );
    vorbis_comment_add_tag( vc, "GENRE",       info->genre   );
    vorbis_comment_add_tag( vc, "YEAR",        info->year    );
    vorbis_comment_add_tag( vc, "DATE",        info->year    );
    vorbis_comment_add_tag( vc, "TRACKNUMBER", info->track   );
    vorbis_comment_add_tag( vc, "COPYRIGHT",   info->album   );
    vorbis_comment_add_tag( vc, "COMMENT",     info->comment );

    if( vcedit_write( state, save ) < 0 ) {
      rc = errno;
    }
    break;
  }

  if( state ) {
    vcedit_clear( state );
  }
  if( file  ) {
    fclose( file );
  }
  if( save  ) {
    fclose( save );
  }

  return rc;
}

/* Returns information about specified track. */
ULONG DLLENTRY
decoder_trackinfo( char* drive, int track, DECODER_INFO* info ) {
  return PLUGIN_NO_PLAY;
}

/* Returns information about a disc inserted to the specified drive. */
ULONG DLLENTRY
decoder_cdinfo( char* drive, DECODER_CDINFO* info ) {
  return PLUGIN_NO_READ;
}

/* What can be played via the decoder. */
ULONG DLLENTRY
decoder_support( char* ext[], int* size )
{
  if( size ) {
    if( ext != NULL && *size > 0 ) {
      strcpy( ext[ 0], "*.OGG" );
    }
    *size = 1;
  }

  return DECODER_FILENAME | DECODER_URL | DECODER_METAINFO;
}

/* Returns information about plug-in. */
void DLLENTRY
plugin_query( PLUGIN_QUERYPARAM* param )
{
  param->type         = PLUGIN_DECODER;
  param->author       = "Dmitry A.Steklenev";
  param->desc         = "OGG Play 2.00";
  param->configurable = FALSE;
}

#if defined(__IBMC__) && defined(__DEBUG_ALLOC__)
unsigned long _System _DLL_InitTerm( unsigned long modhandle,
                                     unsigned long flag )
{
  if( flag == 0 ) {
    if( _CRT_init() == -1 ) {
      return 0UL;
    }
    return 1UL;
  } else {
    _dump_allocated( 0 );
    _CRT_term();
    return 1UL;
  }
}
#endif

