/*
 * Copyright 1997-2003 Samuel Audet <guardia@step.polymtl.ca>
 *                     Taneli Lepp <rosmo@sektori.com>
 *
 * Copyright 2007 Dmitry A.Steklenev <glass@ptv.ru>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *    1. Redistributions of source code must retain the above copyright notice,
 *       this list of conditions and the following disclaimer.
 *
 *    2. Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *
 *    3. The name of the author may not be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


#define  INCL_WIN
#define  INCL_DOS
#include <os2.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <process.h>

#include "pm123.h"
#include "utilfct.h"
#include "pfreq.h"
#include "docking.h"
#include "iniman.h"
#include "snprintf.h"
#include "assertions.h"

static HWND  plman;
static HWND  container;
static HWND  pm_main_menu;
static HWND  pm_list_menu;
static HWND  pm_file_menu;

/* WARNING!! All functions returning a pointer to the
   manager record, return a NULL if suitable record is not found. */

INLINE PMRECORD*
PM_RC( MRESULT rc )
{
  if((PMRECORD*)rc != (PMRECORD*)-1 ) {
    return (PMRECORD*)rc;
  } else {
    return NULL;
  }
}

/* Returns the pointer to the first child record of specified. */
static PMRECORD*
pm_first_child_record( PMRECORD* rec ) {
  return PM_RC( WinSendMsg( container, CM_QUERYRECORD, MPFROMP( rec ),
                            MPFROM2SHORT( CMA_FIRSTCHILD, CMA_ITEMORDER )));
}

/* Returns the pointer to the first playlist manager record. */
static PMRECORD*
pm_first_record( void ) {
  return PM_RC( WinSendMsg( container, CM_QUERYRECORD, NULL,
                            MPFROM2SHORT( CMA_FIRST, CMA_ITEMORDER )));
}

/* Returns the pointer to the next playlist manager record of specified. */
static PMRECORD*
pm_next_record( PMRECORD* rec ) {
  return PM_RC( WinSendMsg( container, CM_QUERYRECORD, MPFROMP(rec),
                            MPFROM2SHORT( CMA_NEXT, CMA_ITEMORDER )));
}

/* Returns the pointer to the record of the specified playlist. */
PMRECORD*
pm_playlist_record( const char* filename )
{
  PMRECORD* rec;

  for( rec = pm_first_record(); rec; rec = pm_next_record( rec )) {
    if( stricmp( filename, rec->filename ) == 0 ) {
      break;
    }
  }
  return rec;
}

/* Returns the pointer to the selected playlist manager record. */
static PMRECORD*
pm_selected( void ) {
  return PM_RC( WinSendMsg( container, CM_QUERYRECORDEMPHASIS,
                            MPFROMP( CMA_FIRST ), MPFROMSHORT( CRA_SELECTED )));
}

/* Frees the data contained in the playlist manager record. */
static void
pm_free_record( PMRECORD* rec )
{
  free( rec->rc.pszTree );
  free( rec->filename   );
}

/* Removes the specified playlist manager record. */
static void
pm_remove_record( PMRECORD* rec )
{
  if( WinSendMsg( container, CM_REMOVERECORD,
                  MPFROMP( &rec ), MPFROM2SHORT( 1, 0 )) != MRFROMLONG( -1 ))
  {
    pm_free_record( rec );
    WinSendMsg( container, CM_FREERECORD, MPFROMP( &rec ), MPFROMSHORT( 1 ));
  }
}

static void
pm_destroy_children_and_self( PMRECORD* rec )
{
  PMRECORD* child;

  while(( child = pm_first_child_record( rec )) != NULL ) {
    pm_remove_record( child );
  }

  pm_remove_record( rec );
}

/* Returns the number of records in the playlist manager. */
static ULONG
pm_size( void )
{
  CNRINFO info;

  if( WinSendMsg( container, CM_QUERYCNRINFO,
                  MPFROMP(&info), MPFROMLONG(sizeof(info))) != 0 )
  {
    return info.cRecords;
  } else {
    return 0;
  }
}

/* Creates the manager record for a specified file. */
PMRECORD*
pm_add_file( const char* filename, PMRECORD* parent, int type,
             int bitrate, int samplerate, int mode, int filesize, int secs )
{
  PMRECORD* rec;
  RECORDINSERT insert;

  rec = (PMRECORD*)WinSendMsg( container, CM_ALLOCRECORD,
                               MPFROMLONG( sizeof( PMRECORD ) - sizeof( RECORDCORE )),
                               MPFROMLONG( 1 ));
  if( rec )
  {
    char title[_MAX_PATH];

    if( type == PM_TYPE_LIST ) {
      sfname( title, filename, sizeof( title ));
    } else {
      if( is_url( filename )) {
        sdecode( title, sfnameext( title, filename, sizeof( title )), sizeof( title ));
      } else {
        sfnameext( title, filename, sizeof( title ));
      }
    }

    rec->rc.cb            = sizeof( RECORDCORE );
    rec->rc.flRecordAttr  = 0;
    rec->rc.pszTree       = strdup( title );
    rec->type             = type;
    rec->filename         = strdup( filename );
    rec->bitrate          = bitrate;
    rec->samplerate       = samplerate;
    rec->mode             = mode;
    rec->filesize         = filesize;
    rec->secs             = secs;
    rec->childs           = 0;

    insert.cb                = sizeof(RECORDINSERT);
    insert.pRecordOrder      = (PRECORDCORE)CMA_END;
    insert.pRecordParent     = (PRECORDCORE)parent;
    insert.fInvalidateRecord = TRUE;
    insert.zOrder            = CMA_TOP;
    insert.cRecordsInsert    = 1;

    if( WinSendMsg( container, CM_INSERTRECORD,
                    MPFROMP( rec ), MPFROMP( &insert )) != 0 )
    {
      if( parent ) {
        if( filesize != -1 ) {
          parent->filesize += filesize;
        }
        if( secs != -1 ) {
          parent->secs += secs;
        }
        ++parent->childs;
      }
    } else {
      pm_free_record( rec );
      WinSendMsg( container, CM_FREERECORD, MPFROMP( &rec ), MPFROMSHORT( 1 ));
      rec = NULL;
    }
  }
  return rec;
}

static void
pm_set_title( const char* title )
{
  static CNRINFO info = { 0 };

  if( info.pszCnrTitle != NULL ) {
    free( info.pszCnrTitle );
  }

  if( title ) {
    info.pszCnrTitle = strdup( title );
    WinSendMsg( container, CM_SETCNRINFO, MPFROMP(&info), MPFROMLONG( CMA_CNRTITLE ));
  } else {
    info.pszCnrTitle = NULL;
  }
}

/* Prints formatted technical data to buffer. */
static char*
pm_sprint_info( char* result, PMRECORD* rec, int size )
{
  char digits[64];

  strlcpy( result, rec->filename, size );

  if( rec->type == PM_TYPE_FILE ) {
    if( rec->secs != -1 ) {
      strlcat( result, ", ", size );
      strlcat( result, itoa( rec->secs / 60, digits, 10 ), size );
      strlcat( result, ":", size );
      strlcat( result, itoa( rec->secs % 60, digits, 10 ), size );
    }
    if( rec->filesize != -1 ) {
      strlcat( result, ", ", size );
      strlcat( result, itoa( rec->filesize / 1024, digits, 10 ), size );
      strlcat( result, " kB", size );
    }
    if( rec->bitrate != -1 ) {
      strlcat( result, ", ", size );
      strlcat( result, itoa( rec->bitrate, digits, 10 ), size );
      strlcat( result, " kb/s", size );
    }
    if( rec->samplerate != -1 ) {
      strlcat( result, ", ", size );
      strlcat( result, itoa( rec->samplerate / 1000, digits, 10 ), size );
      strlcat( result, "." , size );
      strlcat( result, itoa( rec->samplerate % 1000 / 100, digits, 10 ), size );
      strlcat( result, " kHz", size );
    }
    if( rec->mode != -1 ) {
      strlcat( result, ", ", size );
      strlcat( result, modes( rec->mode ), size );
    }
  } else {
    strlcat( result, ", total ", size );
    strlcat( result, itoa( rec->childs, digits, 10 ), size );
    strlcat( result, " files",  size );

    if( rec->filesize > 0 ) {
      strlcat( result, ", ", size );
      strlcat( result, itoa( rec->filesize / 1024, digits, 10 ), size );
      strlcat( result, " kB", size );
    }
    if( rec->secs > 0 ) {
      strlcat( result, ", ", size );
      strlcat( result, itoa( rec->secs / 86400, digits, 10 ), size );
      strlcat( result, "d ", size );
      strlcat( result, itoa( rec->secs % 86400 / 3600, digits, 10 ), size );
      strlcat( result, "h ", size );
      strlcat( result, itoa( rec->secs % 3600 / 60, digits, 10 ), size );
      strlcat( result, "m ", size );
      strlcat( result, itoa( rec->secs % 60, digits, 10 ), size );
      strlcat( result, "s", size );
    }
  }
  return result;
}

static void TFNENTRY
pm_playlist_populate( void* filename )
{
  HAB       hab = NULLHANDLE;
  HMQ       hmq = NULLHANDLE;
  PMRECORD* rec;

  if(( hab = WinInitialize( 0 )) != NULLHANDLE ) {
    hmq = WinCreateMsgQueue( hab, 0 );
  }

  if(( rec = pm_playlist_record( filename )) != NULL ) {
    pl_load( filename, PL_LOAD_TO_PM );

    if( rec == pm_selected()) {
      char title[1024];
      pm_set_title( pm_sprint_info( title, rec, sizeof( title )));
    }
  }
  free( filename );

  if( hmq ) { WinDestroyMsgQueue( hmq ); }
  if( hab ) { WinTerminate( hab ); }
}

static void TFNENTRY
pm_manager_populate( void* dummy )
{
  HAB hab = WinInitialize( 0 );
  HMQ hmq = WinCreateMsgQueue( hab, 0 );

  char  filename[_MAX_PATH];
  char  playlist[_MAX_PATH];
  FILE* file;

  sprintf( filename, "%spm123.mgr", startpath );
  file = fopen( filename, "r" );

  if( file ) {
    while( fgets( playlist, sizeof(playlist), file ))
    {
      blank_strip( playlist );
      if( *playlist && *playlist != '#' && *playlist != '>' && *playlist != '<' ) {
        if( !pm_playlist_record( playlist )) {
          pm_add_file( playlist, NULL, PM_TYPE_LIST, 0, 0, 0, 0, 0 );
          pm_playlist_populate( strdup( playlist ));
        }
      }
    }
    fclose( file );
  }

  WinDestroyMsgQueue( hmq );
  WinTerminate( hab );
}

static void
pm_add_playlist( HWND hwnd )
{
  FILEDLG filedialog;
  char savename[_MAX_PATH];

  int   i       = 0;
  APSZ  types[] = {{ FDT_PLAYLIST }, { 0 }};
  char* file;
  FILE* save;

  memset( &filedialog, 0, sizeof( FILEDLG ));

  filedialog.cbSize         = sizeof( FILEDLG );
  filedialog.fl             = FDS_CENTER | FDS_OPEN_DIALOG | FDS_CUSTOM;
  filedialog.pszTitle       = "Add Playlist(s)";
  filedialog.hMod           = hmodule;
  filedialog.usDlgId        = DLG_FILE;
  filedialog.pfnDlgProc     = amp_file_dlg_proc;
  filedialog.papszITypeList = types;
  filedialog.pszIType       = FDT_PLAYLIST;

  strcpy( filedialog.szFullFile, cfg.listdir );
  WinFileDlg( HWND_DESKTOP, HWND_DESKTOP, &filedialog );

  if( filedialog.lReturn == DID_OK ) {
    if( filedialog.ulFQFCount > 1 ) {
      file = (*filedialog.papszFQFilename)[i];
    } else {
      file = filedialog.szFullFile;
    }

    while( *file ) {
      sdrivedir( cfg.listdir, file, sizeof( cfg.listdir ));
      if( !pm_playlist_record( file )) {
        pm_add_file( file, NULL, PM_TYPE_LIST, 0, 0, 0, 0, 0 );
        _beginthread( pm_playlist_populate, NULL, 65535, strdup( file ));
        sprintf( savename, "%spm123.mgr", startpath );
        if(( save = fopen( savename, "a" )) != NULL ) {
          fprintf( save, "%s\n", file );
          fclose ( save );
        }
      }

      if( ++i >= filedialog.ulFQFCount ) {
        break;
      } else {
        file = (*filedialog.papszFQFilename)[i];
      }
    }
  }

  WinFreeFileDlgList( filedialog.papszFQFilename );
}

static BOOL
pm_remove_playlist( PMRECORD* rec )
{
  char  readname[_MAX_PATH];
  char  savename[_MAX_PATH];
  char  line    [_MAX_PATH];
  FILE* read;
  FILE* save;

  if( !rec || rec->type != PM_TYPE_LIST ) {
    return FALSE;
  }

  sprintf( readname, "%spm123.mgr",    startpath );
  sprintf( savename, "%spm123mgr.$$$", startpath );

  if(( read = fopen( readname, "r" )) == NULL ) {
    return FALSE;
  }
  if(( save = fopen( savename, "w" )) == NULL ) {
    amp_error( plman, "Couldn't open temp file pm123mgr.$$$.\nAre you out of hard disk space?" );
    fclose( read );
    return FALSE;
  }
  while( fgets( line, sizeof(line), read )) {
    blank_strip( line );
    if( stricmp( line, rec->filename ) != 0 ) {
      fprintf( save, "%s\n", line );
    }
  }

  fclose( read );
  fclose( save );

  if( remove( readname ) != 0 ) {
    amp_error( plman, "Unable remove file %s.\n%s\n", readname, strerror( errno ));
    remove( savename );
    return FALSE;
  }
  if( rename( savename, readname ) != 0 ) {
    amp_error( plman, "Unable rename file to %s.\n%s\n", readname, strerror( errno ));
    remove( savename );
    return FALSE;
  }

  pm_destroy_children_and_self( rec );

  WinSendMsg( container, CM_INVALIDATERECORD,
              MPFROMP( NULL ), MPFROM2SHORT( 0, 0 ));

  if( !pm_size()) {
    pm_set_title( "No playlist selected. Right click for menu." );
  }
  return TRUE;
}

/* Shows the context menu of the playlist manager. */
static void
pm_show_context_menu( HWND parent, PMRECORD* rec )
{
  HWND   menu;
  POINTL pos;
  SWP    swp;

  if( rec ) {
    if( rec->type == PM_TYPE_LIST ) {
      menu = pm_list_menu;
    } else {
      menu = pm_file_menu;
    }

    WinSendMsg( container, CM_SETRECORDEMPHASIS,
                MPFROMP( rec ), MPFROM2SHORT( TRUE, CRA_SELECTED | CRA_CURSORED ));
  } else {
    menu = pm_main_menu;
  }

  WinQueryPointerPos( HWND_DESKTOP, &pos );
  WinMapWindowPoints( HWND_DESKTOP, parent, &pos, 1 );

  if( WinWindowFromPoint( parent, &pos, TRUE ) == NULLHANDLE )
  {
    // The context menu is probably activated from the keyboard.
    WinQueryWindowPos( parent, &swp );
    pos.x = swp.cx/2;
    pos.y = swp.cy/2;
  }

  WinPopupMenu( parent, parent, menu, pos.x, pos.y, 0,
                PU_HCONSTRAIN   | PU_VCONSTRAIN | PU_MOUSEBUTTON1 |
                PU_MOUSEBUTTON2 | PU_KEYBOARD   );
}

/* Initializes the playlist manager presentation window. */
static void
pm_init_window( HWND hwnd )
{
  CNRINFO  info;
  HPOINTER hicon;
  LONG     color;
  HACCEL   accel;
  HAB      hab = WinQueryAnchorBlock( hwnd );

  container = WinWindowFromID( hwnd, CNR_PLAYLIST );

  info.flWindowAttr = CV_TREE | CV_TEXT | CA_TREELINE | CA_CONTAINERTITLE | CA_TITLELEFT |
                      CA_TITLESEPARATOR | CA_TITLEREADONLY;
  info.hbmExpanded  = NULLHANDLE;
  info.hbmCollapsed = NULLHANDLE;
  info.cxTreeIndent = 8;
  info.pszCnrTitle  = "No playlist selected. Right click for menu.";

  WinSendMsg( container, CM_SETCNRINFO, MPFROMP( &info ),
              MPFROMLONG( CMA_FLWINDOWATTR | CMA_CXTREEINDENT | CMA_CNRTITLE ));

  accel = WinLoadAccelTable( hab, hmodule, ACL_PM );

  if( accel ) {
    WinSetAccelTable( hab, accel, hwnd );
  }

  hicon = WinLoadPointer( HWND_DESKTOP, hmodule, ICO_MAIN );
  WinSendMsg( hwnd, WM_SETICON, (MPARAM)hicon, 0 );
  do_warpsans( hwnd );

  if( !rest_window_pos( hwnd, 0 ))
  {
    color = CLR_GREEN;
    WinSetPresParam( container, PP_FOREGROUNDCOLORINDEX,
                     sizeof(color), &color);
    color = CLR_BLACK;
    WinSetPresParam( container, PP_BACKGROUNDCOLORINDEX,
                     sizeof(color), &color);
  }
}

/* Processes messages of the playlist manager presentation window. */
static MRESULT EXPENTRY
pm_dlg_proc( HWND hwnd, ULONG msg, MPARAM mp1, MPARAM mp2 )
{
  switch( msg ) {
    case WM_CONTROL:
      switch( SHORT2FROMMP( mp1 )) {
        case CN_HELP:
          amp_show_help( IDH_PM );
          return 0;

        case CN_EMPHASIS:
        {
          PNOTIFYRECORDEMPHASIS emphasis = (PNOTIFYRECORDEMPHASIS)mp2;
          PMRECORD* rec;
          char title[1024];

          if( emphasis != NULL ) {
            rec = (PMRECORD*)emphasis->pRecord;
            pm_set_title( pm_sprint_info( title, rec, sizeof( title )));
          }
          return 0;
        }
        case CN_CONTEXTMENU:
          pm_show_context_menu( hwnd, (PMRECORD*)mp2 );
          return 0;

        case CN_ENTER:
        {
          NOTIFYRECORDENTER* notify = (NOTIFYRECORDENTER*)mp2;
          PMRECORD* rec = (PMRECORD*)notify->pRecord;
          if( rec ) {
            if( rec->type == PM_TYPE_LIST ) {
              pl_load( rec->filename, PL_LOAD_CLEAR );
            } else if( rec->type == PM_TYPE_FILE ) {
              amp_load_singlefile( rec->filename, 0 );
            }
          }
          return 0;
        }
      }
      break;

    case WM_COMMAND:
      switch( COMMANDMSG(&msg)->cmd ) {
        case IDM_PM_CALC:
        {
          int  total_files = 0;
          int  total_secs  = 0;
          int  total_size  = 0;
          char title[1024];

          PMRECORD* rec;

          for( rec = pm_first_record(); rec; rec = pm_next_record( rec )) {
            total_files += rec->childs;
            total_secs  += rec->secs;
            total_size  += rec->filesize;
          }

          snprintf( title, sizeof(title), "Total %u files, %u kB, %ud %uh %um %us",
                    total_files, total_size / 1024, total_secs / 86400, total_secs % 86400 / 3600,
                    total_secs % 3600 / 60,  total_secs % 60 );

          pm_set_title( title );
          break;
        }

        case IDM_PM_L_DELETE:
        {
          PMRECORD* rec = pm_selected();

          if( rec && rec->type == PM_TYPE_LIST ) {
            if( amp_query( plman, "Do you want remove selected playlist from the manager "
                                  "and delete this files from your disk?" ))
            {
              if( pm_remove_playlist( rec )) {
                remove( rec->filename );
              }
            }
          }
          return 0;
        }

        case IDM_PM_L_REMOVE:
          pm_remove_playlist( pm_selected());
          return 0;

        case IDM_PM_L_LOAD:
        {
          PMRECORD* rec = pm_selected();

          if( rec && rec->type == PM_TYPE_LIST ) {
            pl_load( rec->filename, PL_LOAD_CLEAR );
          }
          return 0;
        }

        case IDM_PM_F_LOAD:
        {
          PMRECORD* rec = pm_selected();

          if( rec && rec->type == PM_TYPE_FILE ) {
            amp_load_singlefile( rec->filename, 0 );
          }
          return 0;
        }

        case IDM_PM_ADD:
          pm_add_playlist( hwnd );
          return 0;

        case IDM_PM_MENU:
          WinSendMsg( hwnd, WM_CONTROL,
                      MPFROM2SHORT( CNR_PLAYLIST, CN_CONTEXTMENU ),
                      PVOIDFROMMP( NULL ));
          return 0;

        case IDM_PM_L_MENU:
          WinSendMsg( hwnd, WM_CONTROL,
                      MPFROM2SHORT( CNR_PLAYLIST, CN_CONTEXTMENU ),
                      PVOIDFROMMP( pm_selected()));
          return 0;
      }
      return 0;

    case WM_INITDLG:
      pm_init_window( hwnd );
      dk_add_window ( hwnd, 0 );
     _beginthread( pm_manager_populate, NULL, 65535, NULL );
      return 0;

    case WM_SYSCOMMAND:
      if( SHORT1FROMMP(mp1) == SC_CLOSE ) {
        pm_show( FALSE );
        return 0;
      }
      break;

    case WM_WINDOWPOSCHANGED:
    {
      SWP* pswp = PVOIDFROMMP(mp1);

      if( pswp[0].fl & SWP_SHOW ) {
        cfg.show_plman = TRUE;
      }
      if( pswp[0].fl & SWP_HIDE ) {
        cfg.show_plman = FALSE;
      }
      break;
    }
 }

 return WinDefDlgProc( hwnd, msg, mp1, mp2 );
}

/* Sets the visibility state of the playlist manager presentation window. */
void
pm_show( BOOL show )
{
  HSWITCH hswitch = WinQuerySwitchHandle( plman, 0 );
  SWCNTRL swcntrl;

  if( WinQuerySwitchEntry( hswitch, &swcntrl ) == 0 ) {
    swcntrl.uchVisibility = show ? SWL_VISIBLE : SWL_INVISIBLE;
    WinChangeSwitchEntry( hswitch, &swcntrl );
  }

  dk_set_state( plman, show ? 0 : DK_IS_GHOST );
  WinSetWindowPos( plman, HWND_TOP, 0, 0, 0, 0,
                   show ? SWP_SHOW | SWP_ZORDER | SWP_ACTIVATE : SWP_HIDE );
}

/* Creates the playlist manager presentation window.
   Must be called from the main thread. */
HWND
pm_create( void )
{
  ASSERT_IS_MAIN_THREAD;

  plman = WinLoadDlg( HWND_DESKTOP, HWND_DESKTOP,
                      pm_dlg_proc, hmodule, DLG_PM, NULL );

  pm_list_menu = WinLoadMenu( HWND_OBJECT, hmodule, PM_LIST_MENU );
  pm_file_menu = WinLoadMenu( HWND_OBJECT, hmodule, PM_FILE_MENU );
  pm_main_menu = WinLoadMenu( HWND_OBJECT, hmodule, PM_MAIN_MENU );

  pm_show( cfg.show_plman );
  return plman;
}

/* Destroys the playlist manager presentation window.
   Must be called from the main thread. */
void
pm_destroy( void )
{
  HAB    hab   = WinQueryAnchorBlock( plman );
  HACCEL accel = WinQueryAccelTable ( hab, plman );

  PMRECORD* rec;

  ASSERT_IS_MAIN_THREAD;
  save_window_pos( plman, 0 );

  WinSetAccelTable( hab, plman, NULLHANDLE );
  if( accel != NULLHANDLE ) {
    WinDestroyAccelTable( accel );
  }
  while(( rec = pm_first_record()) != NULL ) {
    pm_destroy_children_and_self( rec );
  }

  pm_set_title( NULL );

  WinDestroyWindow( pm_list_menu );
  WinDestroyWindow( pm_file_menu );
  WinDestroyWindow( pm_main_menu );
  WinDestroyWindow( plman );
}

